/*
 * net/sched/sch_ngwa.c     New Generalized Window Advertising
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Author:	Marcos Talau, <talau@users.sourceforge.net>
 *
 * UPDATES: The latest version of this source code can be obtained by the above address.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/bitops.h>
#include <linux/list.h>
#include <net/pkt_sched.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <asm/byteorder.h>

/* active DPF? */
static int dpf = 0;
module_param (dpf, bool, S_IRUGO);
MODULE_PARM_DESC (dpf, "true to use NGWA DPF at queue");
static int dpf_aa = 10;
module_param (dpf_aa, int, S_IRUGO);
MODULE_PARM_DESC (dpf_aa, "NGWA DPF aa (average ack)");
static int dpf_hs = 20;
module_param (dpf_hs, int, S_IRUGO);
MODULE_PARM_DESC (dpf_hs, "NGWA DPF hs (header size)");

extern int sch_total_pkt_size;
extern int sch_total_pkt;

struct ngwa_tcp_connection {
	__be32 saddr;
	__be32 daddr;
	__be16 sport;
	__be16 dport;
	__u8 protocol;
	struct list_head list;
};
struct ngwa_tcp_connection *ngwa_tcp_conn;
static LIST_HEAD(ngwa_tcp_conn_list);

struct ngwa_sched_data
{
	u32 limit;
	struct Qdisc		*q;
};

// used only for dpf param; the number of active connections between queue
static int dpf_conn = 0;

// return average queue len free in bytes
static u32 ngwa_get_available_bandwidth(struct Qdisc *sch)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	u32 mean_pkt_size;
	u32 bytes_free;

	// with out (u32) the variable bstats.bytes use 64 bits,
	// so in 32 bits compilation the error '__udivdi3' will appear
	mean_pkt_size = (u32) sch_total_pkt_size / sch_total_pkt;
	bytes_free = (q->limit - sch->q.qlen) * mean_pkt_size;

	if (dpf && dpf_conn && (bytes_free > dpf_conn)) {
		printk(KERN_DEBUG "(ngwa_get_available_bandwidth) (dpf) total bytes_free = %d\n", bytes_free);
		bytes_free /= dpf_conn;
		printk(KERN_DEBUG "(ngwa_get_available_bandwidth) (dpf) with %d conn bytes_free = %d\n", dpf_conn, bytes_free);
		bytes_free = bytes_free - dpf_aa * (dpf_hs * dpf_conn);
		printk(KERN_DEBUG "(ngwa_get_available_bandwidth) (dpf) with %d conn bytes_free (with aa=%d and hs=%d) = %d\n", dpf_conn, dpf_aa, dpf_hs, bytes_free);
	}

	return bytes_free;
}

static u32 ngwa_get_ngwa_pkt(struct sk_buff *skb)
{
	struct iphdr *iph;
	struct ip_options *opt;
	// will point to ip->opt real data
	unsigned char *sptr;

	iph = ip_hdr(skb);
	opt = &(IPCB(skb)->opt);

	if (opt && opt->optlen) {
		// ngwa are in __pad2 pointer
		if (opt->__pad2) {
			sptr = skb_network_header(skb);
	
			if (sptr[opt->__pad2] == IPOPT_NGWA) {
				printk(KERN_DEBUG "(ngwa_get_ngwa_pkt) ngwa in received skb had %d\n", (u16) ((u16) sptr[opt->__pad2+3] << 8) | (u16) sptr[opt->__pad2+2]);
				// 8 bits from __pad2+2 and 8 too from __pad2+3 = 16 bits
				return (u16) ((u16) sptr[opt->__pad2+3] << 8) | (u16) sptr[opt->__pad2+2];
			}
		}
	}
	// if skb do not have ngwa already
	return 0;
}

static void ngwa_update_pkt(struct Qdisc *sch, struct sk_buff *skb, u32 ngwa_bw)
{
	// optlen in bytes
	u8 optlen = NGWA_OLEN;
	struct iphdr *iph = ip_hdr(skb);

	__be32 daddr = 0; 
	struct ip_options *old_opt;
	struct ip_options *new_opt;
	// real because it have 16 bits
	u16 real_bw;
	// we will alloc ipopt? if ip already have ip->opt we don't
	u8 alloc_ipopt = 0;

	real_bw = min(ngwa_bw, MAX_TCP_WINDOW);

	// check if already have ipopt
	old_opt = &(IPCB(skb)->opt);
	if (!(old_opt && old_opt->optlen)) {
		alloc_ipopt = 1;

		iph = ip_hdr(skb);
		if (iph->ihl != 5)
			printk(KERN_EMERG "hmm unexpected ihl len. FIXME\n");

		pskb_allocate_ip_options(skb, optlen, GFP_ATOMIC);
	}

	iph = ip_hdr(skb);

	new_opt = kmalloc(sizeof (struct ip_options) + 40, GFP_KERNEL);
	memset(new_opt, 0, sizeof(struct ip_options) + 40);
	new_opt->__data[0] = IPOPT_NGWA; // type of option
	new_opt->__data[1] = optlen; // optlen
	// here [2]/[3] we split real_bw into two parts with 8 bits
	// [2] = right part, [3] = left part
	new_opt->__data[2] = real_bw & 0xff;
	new_opt->__data[3] = (real_bw >> 8) & 0xff;
	new_opt->__data[4] = IPOPT_NOOP;
	new_opt->__data[5] = IPOPT_NOOP;
	new_opt->__data[6] = IPOPT_NOOP;
	new_opt->__data[7] = IPOPT_END;
	new_opt->optlen = optlen;
	if (alloc_ipopt)
		iph->ihl += new_opt->optlen>>2;

	ip_options_build(skb, new_opt, daddr, NULL, 0);
	iph = ip_hdr(skb);

	if (alloc_ipopt) {
		iph->tot_len += htons(optlen);
		skb->len += optlen;
	}

	// add checksum
	ip_send_check(iph);
	atomic_set(&skb_shinfo(skb)->dataref, 1);

	old_opt = &(IPCB(skb)->opt);
	if (!(old_opt && old_opt->optlen)) {
		printk(KERN_EMERG "Oh shit! ngwa do not set options\n");
	}

	if (skb->len > 1518)
		printk(KERN_EMERG "NGWA problem. Packet (skb) size is too large. hint: Check if MSS in TCP sides are with - 8?\n");

	printk(KERN_DEBUG "(ngwa_update_pkt) updated to %d\n", real_bw);
}

static void ngwa_add_conn(struct iphdr *iph, struct tcphdr *tcph)
{
	struct ngwa_tcp_connection *new;

	new = kmalloc(sizeof(*new), GFP_KERNEL);
	new->saddr = iph->saddr;
	new->daddr = iph->daddr;
	new->sport = tcph->source;
	new->dport = tcph->dest;
	new->protocol = iph->protocol;
	list_add(&new->list, &ngwa_tcp_conn_list);
	dpf_conn++;

	printk(KERN_EMERG "(ngwa_add_conn) Now number of connections is %d.\n", dpf_conn);
}

static struct ngwa_tcp_connection * ngwa_find_conn(struct iphdr *iph, struct tcphdr *tcph)
{
	struct ngwa_tcp_connection *conn;

	list_for_each_entry(conn, &ngwa_tcp_conn_list, list) {
		if ( (iph->saddr == conn->saddr) && (iph->daddr == conn->daddr)
			&& (tcph->source == conn->sport) && (tcph->dest == conn->dport)
			&& (iph->protocol == conn->protocol))
			return conn;
		// two side check
		if ( (iph->saddr == conn->daddr) && (iph->daddr == conn->saddr)
			&& (tcph->source == conn->dport) && (tcph->dest == conn->sport)
			&& (iph->protocol == conn->protocol))
			return conn;
	}

	return NULL;
}

static void ngwa_del_conn(struct ngwa_tcp_connection *conn)
{
	if (likely(conn)) {
		printk(KERN_DEBUG "(ngwa_del_conn) conn will be deleted\n");
		list_del(&conn->list);
	} else {
		printk(KERN_DEBUG "----- conn NOT will be deleted\n");
	}

	if (dpf_conn == 0)
		return;
	dpf_conn--;
	printk(KERN_EMERG "(ngwa_del_conn) Now number of connections is %d.\n", dpf_conn);
}

// main ngwa cicle in queue
static void ngwa_process(struct Qdisc *sch, struct sk_buff *skb)
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *tcph;
	u32 ngwa_local_bw;
	u32 ngwa_from_pkt;
	struct ip_options *opt;
	u8 ip_len = 20;

	// ngwa will apply only in TCP protocol
	if ((iph->frag_off & htons(IP_OFFSET) ? 0 : iph->protocol) != IPPROTO_TCP)
		return;

	if (dpf) {
		// check if already have ipopt
		opt = &(IPCB(skb)->opt);

		if (opt && opt->optlen)
			ip_len += NGWA_OLEN;
			
	// always keep transport_header correct (when forward a packet transport_header will not be adjusted)
#ifdef NET_SKBUFF_DATA_USES_OFFSET
		skb->transport_header = skb->network_header + ip_len;
#else
		skb->transport_header = skb->head + (skb->network_header - skb->head) + ip_len;
#endif
		tcph = tcp_hdr(skb);

		printk(KERN_DEBUG "(ngwa_process)(dpf) DPF is on ;; active tcp conn = %d\n", dpf_conn);
		printk(KERN_DEBUG "(ngwa_process)(dpf) out-int %s with ip pkt from: %pI4 to: %pI4 ;; flags: fin %d, ack %d, psh %d, syn %d\n", skb->dev->name, &iph->saddr, &iph->daddr, tcph->fin, tcph->ack, tcph->psh, tcph->syn);
		// DPF part
		// here only for SYN+ACK
		if (tcph->syn) {
			printk(KERN_DEBUG "(ngwa_process)(dpf) syn/? received ;; (%pI4 => %pI4) flags: fin %d, ack %d, psh %d, syn %d\n", &iph->saddr, &iph->daddr, tcph->fin, tcph->ack, tcph->psh, tcph->syn);
			if (! ngwa_find_conn(iph, tcph)) {
				printk(KERN_DEBUG "(ngwa_process)(dpf) new conn will be registred\n");
				ngwa_add_conn(iph, tcph);
			}
		} else if (tcph->fin || tcph->rst) {
			printk(KERN_DEBUG "(ngwa_process)(dpf) fin or rst received\n");
			ngwa_del_conn(ngwa_find_conn(iph, tcph));
		}
	}

	ngwa_local_bw = ngwa_get_available_bandwidth(sch);
	ngwa_from_pkt = ngwa_get_ngwa_pkt(skb);

	printk(KERN_DEBUG "(ngwa_process) ngwa_local %d; ngwa_from_pkt %d\n", ngwa_local_bw, ngwa_from_pkt);

	if ((ngwa_local_bw < ngwa_from_pkt) || (ngwa_from_pkt == 0)) {
		printk(KERN_DEBUG "(ngwa_process) I will update pkt due less bw.\n");
		if (likely(ngwa_local_bw))
			ngwa_update_pkt(sch, skb, ngwa_local_bw);
	}
}

// only for debug process
static void show_info(struct sk_buff *skb)
{
	struct iphdr *iph = ip_hdr(skb);

	printk(KERN_DEBUG "[sch_ngwa](SHOWINFO) [int %s]; ip version %d; from: %pI4 to: %pI4\n", skb->dev->name, iph->version, &iph->saddr, &iph->daddr);
}

static int ngwa_enqueue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	int ret;

	ret = qdisc_enqueue(skb, q->q);
	if (ret != NET_XMIT_SUCCESS) {
		if (net_xmit_drop_count(ret))
			sch->qstats.drops++;
		return ret;
	}

	sch->bstats.bytes += qdisc_pkt_len(skb);
	sch->bstats.packets++;
	sch->q.qlen++;

	return NET_XMIT_SUCCESS;
}

static struct sk_buff * ngwa_dequeue(struct Qdisc* sch)
{
	struct sk_buff *skb;
	struct ngwa_sched_data *q = qdisc_priv(sch);

	skb = q->q->ops->dequeue(q->q);
	if (skb == NULL)
		return NULL;

	sch_total_pkt_size += qdisc_pkt_len(skb);
	if (sch_total_pkt == 0)
		sch_total_pkt = 1;

	sch->q.qlen--;

	if (skb)
		ngwa_process(sch, skb);

	return skb;
}

static int ngwa_init(struct Qdisc *sch, struct nlattr *opt)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	u32 limit = qdisc_dev(sch)->tx_queue_len ? : 1;

	q->limit = limit;
	sch->q.qlen = 0;
	sch->bstats.bytes = 0;
	sch->bstats.packets = 0;

	if (dpf)
		printk(KERN_EMERG "(ngwa_init) DPF is active.\n");

	printk(KERN_EMERG "(ngwa_init) sch_total_pkt_size restarted.\n");
	sch_total_pkt_size = 0;
	sch_total_pkt = 0;

	q->q = qdisc_create_dflt(qdisc_dev(sch), sch->dev_queue,
				 &pfifo_qdisc_ops, sch->handle);
	if (q->q == NULL)
		q->q = &noop_qdisc;

	ngwa_tcp_conn = kmalloc(sizeof(*ngwa_tcp_conn), GFP_KERNEL);
	INIT_LIST_HEAD(&ngwa_tcp_conn->list);

	return 0;
}

static int ngwa_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	struct tc_fifo_qopt opt = { .limit = q->limit };

	NLA_PUT(skb, TCA_OPTIONS, sizeof(opt), &opt);
	return skb->len;

nla_put_failure:
	return -1;
}

static unsigned int ngwa_drop(struct Qdisc *sch)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	unsigned int len;

	if (q->q->ops->drop == NULL)
		return 0;

	len = q->q->ops->drop(q->q);
	if (len)
		sch->q.qlen--;

	return len;
}

static struct sk_buff *ngwa_peek(struct Qdisc *sch)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);

	return q->q->ops->peek(q->q);
}

static void ngwa_reset(struct Qdisc *sch)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);

	qdisc_reset(q->q);
	sch->q.qlen = 0;
	sch->bstats.bytes = 0;
	sch->bstats.packets = 0;
}

static int ngwa_graft(struct Qdisc *sch, unsigned long arg,
			struct Qdisc *new, struct Qdisc **old)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);

	if (new == NULL) {
		new = qdisc_create_dflt(qdisc_dev(sch), sch->dev_queue,
					&pfifo_qdisc_ops,
					sch->handle);
		if (new == NULL)
			new = &noop_qdisc;
	}

	sch_tree_lock(sch);
	*old = q->q;
	q->q = new;
	qdisc_tree_decrease_qlen(*old, (*old)->q.qlen);
	qdisc_reset(*old);
	sch_tree_unlock(sch);

	return 0;
}

static struct Qdisc *ngwa_leaf(struct Qdisc *sch, unsigned long arg)
{
	struct ngwa_sched_data *q = qdisc_priv(sch);
	return q->q;
}

static unsigned long ngwa_get(struct Qdisc *sch, u32 classid)
{
	return TC_H_MIN(classid) + 1;
}

static void ngwa_walk(struct Qdisc *sch, struct qdisc_walker *walker)
{
	if (!walker->stop) {
		if (walker->count >= walker->skip)
			if (walker->fn(sch, 1, walker) < 0) {
				walker->stop = 1;
				return;
			}
		walker->count++;
	}
}

static void ngwa_put(struct Qdisc *sch, unsigned long arg)
{
	return;
}

static const struct Qdisc_class_ops ngwa_class_ops = {
	.graft		=	ngwa_graft,
	.leaf		=	ngwa_leaf,
	.get		=	ngwa_get,
	.put		=	ngwa_put,
	.walk		=	ngwa_walk,
};

static struct Qdisc_ops ngwa_qdisc_ops __read_mostly = {
	.id		=	"ngwa",
	.priv_size	=	sizeof(struct ngwa_sched_data),
	.cl_ops		=	&ngwa_class_ops,
	.enqueue	=	ngwa_enqueue,
	.dequeue	=	ngwa_dequeue,
	.peek		=	ngwa_peek,
	.drop		=	ngwa_drop,
	.init		=	ngwa_init,
	.reset		=	ngwa_reset,
	.change		=	ngwa_init, // it can change
	.dump		=	ngwa_dump,
	.owner		=	THIS_MODULE,
};

static int __init ngwa_module_init(void)
{
	return register_qdisc(&ngwa_qdisc_ops);
}

static void __exit ngwa_module_exit(void)
{
	unregister_qdisc(&ngwa_qdisc_ops);
}

module_init(ngwa_module_init)
module_exit(ngwa_module_exit)

MODULE_LICENSE("GPL");
